module ConvectionTools
using Dash, DashBootstrapComponents
using PlotlyJS
using LaMEM
using UUIDs
using Interpolations
using GeophysicalModelGenerator
using DelimitedFiles
using HTTP

export convection 

pkg_dir = Base.pkgdir(ConvectionTools)

include(joinpath(pkg_dir,"src/dash_tools.jl"))
include(joinpath(pkg_dir,"src/RayleighBenardConvection/dash_functions_convection.jl"))
 

"""
    convection(; host=HTTP.Sockets.localhost, port=8050)

This starts a convection GUI
"""
function convection(; host=HTTP.Sockets.localhost, port=8050, width="80vw", height="60vh")
    pkg_dir = Base.pkgdir(ConvectionTools)
    cmaps = read_colormaps(dir_colormaps=joinpath(pkg_dir,"src/assets/colormaps/"))
    
    GUI_version = "0.1.3"
    
    title_app = "Rayleigh-Benard convection"
    ParamFile = "Convection.dat"
    OutFile = "Convection"

    base_dir = pwd()
    cd(base_dir)

    #app = dash(external_stylesheets=[dbc_themes.CYBORG])
    app = dash(external_stylesheets = [dbc_themes.BOOTSTRAP, dbc_icons.BOOTSTRAP], prevent_initial_callbacks=false)
    app.title = title_app

    # Main code layout
    app.layout = html_div() do
        dbc_container(className="mxy-auto", fluid=true, [
            make_title(title_app),
            dbc_row([
                dbc_col([
                    make_plot("",cmaps, width=width, height=height),    # show graph
                    make_plot_controls(),   # show media buttons
                    make_id_label(),        # show user id
                ]),
                dbc_col([
                    make_time_card(),                               # show simulation time info
                    make_menu(cmaps, show_field="temperature [°C]"),     # show menu with simulation parameters, rheological parameters, and plotting parameters
                    make_run_button()                               # show the run simulation button
                ])
            ]),

            # Store a unique number of our session in the webpage
            dcc_store(id="session-id", data=""),

            # Store info related to the simulation and current timestep
            dcc_store(id="current_timestep", data="0"),
            dcc_store(id="last_timestep", data="0"),
            dcc_store(id="update_fig", data="0"),

            # Start an interval that updates the number every second
            dcc_interval(id="session-interval", interval=100, n_intervals=0, disabled=true)
        ])

    end

    # This creates an initial session id that is unique for this session
    # it will run on first start 
    callback!(app, 
        Dash.Output("session-id", "data"),
        Dash.Output("label-id", "children"),
        Input("session-id", "data")
    ) do session_id

        session_id = UUIDs.uuid4()
        str = "id=$(session_id), v=$(GUI_version)"
        return String("$(session_id)"), str
    end

    # Call run button
    callback!(app,
        Dash.Output("session-interval", "disabled"),
        Input("button-run", "n_clicks"),
        Input("button-run", "disabled"),
        Input("button-play", "n_clicks"), 
        State("domain_width", "value"),
        State("domain_height", "value"),
        State("nel_x", "value"),
        State("nel_z", "value"),
        State("n_timesteps", "value"),
        State("ΔT", "value"),
        State("γ", "value"),
        State("cohesion", "value"),
        State("viscosity", "value"),
        State("last_timestep", "data"),
        State("plot_field", "value"),
        State("session-id", "data"),
        State("switch-FreeSurf","value"),
        prevent_initial_call=true
    ) do n_run, active_run, n_play,
        domain_width, domain_height, nel_x, nel_z, n_timesteps,
        ΔT, γ, cohesion, viscosity,
        last_timestep, plot_field, session_id, FreeSurface
        
        trigger = get_trigger()
        disable_interval = true
        user_dir = simulation_directory(session_id, clean=false)
        if trigger == "button-run.n_clicks"
            cd(pkg_dir)
            cur_dir = pwd()
            base_dir = joinpath(pkgdir(ConvectionTools),"src","RayleighBenardConvection")

            viscosity = 10.0^viscosity
            cohesion *= 1.0e6

            Δx  = domain_width/nel_x # y-width
            
            if FreeSurface === nothing
                args = "-nstep_max $(n_timesteps) -eta_fk[0] $(viscosity)  -gamma_fk[0] $γ -TRef_fk[0] $(ΔT/2) -ch[0] $(cohesion) -nel_x $nel_x -nel_z $nel_z -coord_x $(-domain_width/2),$(domain_width/2) -coord_z $(-domain_height),0 -coord_y $(-Δx/2),$(Δx/2) -temp_bot $ΔT"  
            else
                args = "-nstep_max $(n_timesteps) -eta_fk[0] $(viscosity)  -gamma_fk[0] $γ -TRef_fk[0] $(ΔT/2) -ch[0] $(cohesion) -nel_x $nel_x -coord_x $(-domain_width/2),$(domain_width/2) -coord_y $(-Δx/2),$(Δx/2) -temp_bot $ΔT"
            end

            println("args = ", args)

            # We clicked the run button
            user_dir = simulation_directory(session_id, clean=true)
            cd(user_dir)
            
            pfile = joinpath(base_dir,ParamFile) 
            CreateSetup(pfile, ΔT, args=args)
            run_lamem(pfile, 1, args, wait=false)
            disable_interval = false
            cd(cur_dir)        # return to main directory

        elseif trigger == "button-run.disabled"
            last_t = parse(Int, last_timestep)
            if active_run == true || last_t < n_timesteps
                disable_interval = false
            end

        elseif trigger == "button-play.n_clicks"
            last_t = parse(Int, last_timestep)
            # @show last_t
            disable_interval = false
        end
        
        return disable_interval
    end

    
    # deactivate the button 
    callback!(app,
        Dash.Output("button-run", "disabled"),
        Dash.Output("button-run", "color"),
        Input("button-run", "n_clicks"),
        Input("session-interval", "n_intervals"),
        State("last_timestep", "data"),
        State("current_timestep", "data"),
        prevent_initial_call=true
    ) do n_run, n_inter, last_timestep, current_timestep

        cur_t = parse(Int, current_timestep)    # current timestep
        last_t = parse(Int, last_timestep)      # last timestep available on disk
        if cur_t < last_t
            button_run_disable = true
            button_color = "danger"
        else
            button_run_disable = false
            button_color = "primary"
        end

        return button_run_disable, button_color
    end

    # Check if *.pvd file on disk changed and a new timestep is available
    callback!(app,
        Dash.Output("last_timestep", "data"),
        Dash.Output("update_fig", "data"),
        Input("session-interval", "n_intervals"),
        Input("button-run", "n_clicks"),
        State("current_timestep", "data"),
        State("update_fig", "data"),
        State("session-id", "data"),
        prevent_initial_call=true
    ) do n_inter, n_run, current_timestep, update_fig, session_id
        trigger = get_trigger()
        user_dir = simulation_directory(session_id, clean=false)
        if trigger == "session-interval.n_intervals"
            if has_pvd_file(OutFile, user_dir)
                # Read LaMEM *.pvd file
                Timestep, _, Time = read_LaMEM_simulation(OutFile, user_dir)

                # Update the labels and data stored in webpage about the last timestep
                last_time = "$(Timestep[end])"

                update_fig = "$(parse(Int,update_fig)+1)"
            else
                last_time = "0"
                update_fig = "0"
            end
        elseif trigger == "button-run.n_clicks"

            last_time = "0"
            update_fig = "0"
        end

        return last_time, update_fig
    end

    # Update the figure if the signal is given to do so
    callback!(app,
        Dash.Output("label-timestep", "children"),
        Dash.Output("label-time", "children"),
        Dash.Output("current_timestep", "data"),
        Dash.Output("figure_main", "figure"),
        Dash.Output("plot_field", "options"),
        Dash.Output("contour_option", "options"),
        Input("update_fig", "data"),
        Input("current_timestep", "data"),
        Input("button-run", "n_clicks"),
        Input("button-start", "n_clicks"),
        Input("button-last", "n_clicks"),
        Input("button-forward", "n_clicks"),
        Input("button-back", "n_clicks"),
        Input("button-play", "n_clicks"),
        State("last_timestep", "data"),
        State("session-id", "data"),
        State("plot_field", "value"),
        State("switch-contour", "value"),
        State("contour_option", "value"),
        State("switch-velocity", "value"),
        State("color_map_option", "value"),
        prevent_initial_call=true
    ) do update_fig, current_timestep, n_run, n_start, n_last, n_back, n_forward, n_play, last_timestep, session_id,
    plot_field, switch_contour, contour_field, switch_velocity, color_map_option

        trigger = get_trigger()

        # Get info about timesteps
        cur_t = parse(Int, current_timestep)                    # current timestep
        last_t = parse(Int, last_timestep)                      # last timestep available on disk
        fig_cross = []
        fields_available = ["phase"]
        if trigger == "current_timestep.data" ||
        trigger == "update_fig.data" ||
        trigger == "button-start.n_clicks" ||
        trigger == "button-last.n_clicks" ||
        trigger == "button-back.n_clicks" ||
        trigger == "button-forward.n_clicks" ||
        trigger == "button-play.n_clicks"

            user_dir = simulation_directory(session_id, clean=false)
            if  has_pvd_file(OutFile, user_dir)
                Timestep, _, Time = read_LaMEM_simulation(OutFile, user_dir)      # all timesteps
                id = findall(Timestep .== cur_t)[1]
                if trigger == "button-start.n_clicks" || trigger == "button-play.n_clicks"
                    cur_t = 0
                    id = 1
                elseif trigger == "button-last.n_clicks"
                    cur_t = Timestep[end]
                    id = length(Timestep)
                elseif (trigger == "button-forward.n_clicks") && (id < length(Timestep))
                    cur_t = Timestep[id+1]
                    id = id + 1
                elseif (trigger == "button-back.n_clicks") && (id > 1)
                    cur_t = Timestep[id-1]
                    id = id - 1
                end

                # Load data 
                x, y, data, time, fields_available = get_data(OutFile, cur_t, plot_field, user_dir)
                add_contours = active_switch(switch_contour)
                if add_contours
                    x_con, y_con, data_con, _, _ = get_data(OutFile, cur_t, contour_field, user_dir)
                else
                    x_con, y_con, data_con = x, y, data
                end
                
                # update the plot
                add_velocity = active_switch(switch_velocity)

                fig_cross = create_main_figure(OutFile, cur_t, x, y, data, x_con, y_con, data_con;
                    add_contours=add_contours, contour_field=contour_field,
                    add_velocity=add_velocity,
                    colorscale=color_map_option,
                    session_id=session_id, 
                    cmaps=cmaps,
                    field=plot_field)

                if trigger == "current_timestep.data" || trigger == "update_fig.data" || trigger == "button-play.n_clicks"
                    if cur_t < last_t
                        cur_t = Timestep[id+1]      # update current timestep
                    end
                end

            else
                time = 0
            end

        elseif trigger == "button-run.n_clicks"
            cur_t = 0
            time = 0.0
        end

        # update the labels
        label_timestep = "Timestep: $cur_t"
        label_time = "Time: $time Myrs"
        current_timestep = "$cur_t"

        # @show current_timestep
        println("Timestep ", cur_t)
        return label_timestep, label_time, current_timestep, fig_cross, add_units(fields_available), add_units(fields_available)
    end

    # 
    callback!(app,
        Dash.Output("contour_option", "disabled"),
        Input("switch-contour", "value")) do switch_contour
        if !isnothing(switch_contour)
            if isempty(switch_contour)
                disable_contours = true
            else
                disable_contours = false
            end
        else
            disable_contours = true
        end
        return disable_contours
    end
    

    run_server(app, host, port, debug=false)


    return app
end

end
